@@ -47,19 +47,21 @@
 #include "propertydeclaration.h"
 #include "value.h"
 #include <logging/translator.h>
-#include <tools/architectures.h>
 #include <tools/fileinfo.h>
-#include <tools/hostosinfo.h>
 #include <tools/qbsassert.h>
+#include <tools/qttools.h>
 #include <tools/scripttools.h>
-#include <tools/shellutils.h>
-#include <jsextensions/environmentextension.h>
+#include <tools/stringconstants.h>
 
-#include <QByteArray>
-#include <QScriptString>
-#include <QScriptValue>
-#include <QDebug>
-#include <QSettings>
+#include <QtCore/qbytearray.h>
+#include <QtCore/qdebug.h>
+#include <QtCore/qsettings.h>
+
+#include <QtScript/qscriptclasspropertyiterator.h>
+#include <QtScript/qscriptstring.h>
+#include <QtScript/qscriptvalue.h>
+
+#include <utility>
 
 namespace qbs {
 namespace Internal {
@@ -70,29 +72,27 @@ class SVConverter : ValueHandler
     ScriptEngine * const engine;
     QScriptContext * const scriptContext;
     const QScriptValue * const object;
-    const ValuePtr &valuePtr;
+    Value * const valuePtr;
     const Item * const itemOfProperty;
     const QScriptString * const propertyName;
     const EvaluationData * const data;
     QScriptValue * const result;
-    QStack<JSSourceValue *> * const sourceValueStack;
     char pushedScopesCount;
 
 public:
 
     SVConverter(EvaluatorScriptClass *esc, const QScriptValue *obj, const ValuePtr &v,
                 const Item *_itemOfProperty, const QScriptString *propertyName, const EvaluationData *data,
-                QScriptValue *result, QStack<JSSourceValue *> *sourceValueStack)
+                QScriptValue *result)
         : scriptClass(esc)
         , engine(static_cast<ScriptEngine *>(esc->engine()))
         , scriptContext(esc->engine()->currentContext())
         , object(obj)
-        , valuePtr(v)
+        , valuePtr(v.get())
         , itemOfProperty(_itemOfProperty)
         , propertyName(propertyName)
         , data(data)
         , result(result)
-        , sourceValueStack(sourceValueStack)
         , pushedScopesCount(0)
     {
     }
@@ -103,11 +103,30 @@ public:
     }
 
 private:
+    friend class AutoScopePopper;
+
+    class AutoScopePopper
+    {
+    public:
+        AutoScopePopper(SVConverter *converter)
+            : m_converter(converter)
+        {
+        }
+
+        ~AutoScopePopper()
+        {
+            m_converter->popScopes();
+        }
+
+    private:
+        SVConverter *m_converter;
+    };
+
     void setupConvenienceProperty(const QString &conveniencePropertyName, QScriptValue *extraScope,
                                   const QScriptValue &scriptValue)
     {
         if (!extraScope->isObject())
-            *extraScope = scriptClass->engine()->newObject();
+            *extraScope = engine->newObject();
         const PropertyDeclaration::Type type
                 = itemOfProperty->propertyDeclaration(propertyName->toString()).type();
         const bool isArray = type == PropertyDeclaration::StringList
@@ -123,6 +142,60 @@ private:
         extraScope->setProperty(conveniencePropertyName, valueToSet);
     }
 
+    std::pair<QScriptValue, bool> createExtraScope(const JSSourceValue *value, Item *outerItem,
+                                                   QScriptValue *outerScriptValue)
+    {
+        std::pair<QScriptValue, bool> result;
+        auto &extraScope = result.first;
+        result.second = true;
+        if (value->sourceUsesBase()) {
+            QScriptValue baseValue;
+            if (value->baseValue()) {
+                SVConverter converter(scriptClass, object, value->baseValue(), itemOfProperty,
+                                      propertyName, data, &baseValue);
+                converter.start();
+            }
+            setupConvenienceProperty(StringConstants::baseVar(), &extraScope, baseValue);
+        }
+        if (value->sourceUsesOuter()) {
+            QScriptValue v;
+            if (outerItem) {
+                v = data->evaluator->property(outerItem, *propertyName);
+                if (engine->hasErrorOrException(v)) {
+                    extraScope = engine->lastErrorValue(v);
+                    result.second = false;
+                    return result;
+                }
+            } else if (outerScriptValue) {
+                v = *outerScriptValue;
+            }
+            if (v.isValid())
+                setupConvenienceProperty(StringConstants::outerVar(), &extraScope, v);
+        }
+        if (value->sourceUsesOriginal()) {
+            QScriptValue originalValue;
+            if (data->item->propertyDeclaration(propertyName->toString()).isScalar()) {
+                const Item *item = itemOfProperty;
+                while (item->type() == ItemType::ModuleInstance)
+                    item = item->prototype();
+                if (item->type() != ItemType::Module && item->type() != ItemType::Export) {
+                    const QString errorMessage = Tr::tr("The special value 'original' can only "
+                                                        "be used with module properties.");
+                    extraScope = engine->currentContext()->throwError(errorMessage);
+                    result.second = false;
+                    return result;
+                }
+                SVConverter converter(scriptClass, object, item->property(*propertyName), item,
+                                      propertyName, data, &originalValue);
+                converter.start();
+            } else {
+                originalValue = engine->newArray(0);
+            }
+            setupConvenienceProperty(StringConstants::originalVar(), &extraScope, originalValue);
+        }
+        return result;
+    }
+
     void pushScope(const QScriptValue &scope)
     {
         if (scope.isObject()) {
@@ -146,145 +219,117 @@ private:
             scriptContext->popScope();
     }
 
-    void handle(JSSourceValue *value)
+    void handle(JSSourceValue *value) override
     {
-        const Item *conditionScopeItem = 0;
-        QScriptValue conditionScope;
-        QScriptValue conditionFileScope;
-        Item *outerItem = data->item->outerItem();
-        for (int i = 0; i < value->alternatives().count(); ++i) {
-            const JSSourceValue::Alternative *alternative = 0;
-            alternative = &value->alternatives().at(i);
-            if (!conditionScopeItem) {
-                // We have to differentiate between module instances and normal items here.
-                //
-                // The module instance case:
-                // Product {
-                //     property bool something: true
-                //     Properties {
-                //         condition: something
-                //         cpp.defines: ["ABC"]
-                //     }
-                // }
-                //
-                // data->item points to cpp and the condition's scope chain must contain cpp's
-                // scope, which is the item where cpp is instantiated. The scope chain must not
-                // contain data->item itself.
-                //
-                // The normal item case:
-                // Product {
-                //     property bool something: true
-                //     property string value: "ABC"
-                //     Properties {
-                //         condition: something
-                //         value: "DEF"
-                //     }
-                // }
-                //
-                // data->item points to the product and the condition's scope chain must contain
-                // the product item.
-                conditionScopeItem = data->item->type() == ItemType::ModuleInstance
-                        ? data->item->scope() : data->item;
-                conditionScope = data->evaluator->scriptValue(conditionScopeItem);
-                QBS_ASSERT(conditionScope.isObject(), return);
-                conditionFileScope = data->evaluator->fileScope(value->file());
-            }
-            engine->currentContext()->pushScope(conditionFileScope);
-            pushItemScopes(conditionScopeItem);
-            if (alternative->value->definingItem())
-                pushItemScopes(alternative->value->definingItem());
-            engine->currentContext()->pushScope(conditionScope);
-            engine->currentContext()->pushScope(data->evaluator->importScope(value->file()));
-            const QScriptValue cr = engine->evaluate(alternative->condition);
-            const QScriptValue overrides = engine->evaluate(alternative->overrideListProperties);
-            engine->currentContext()->popScope();
-            engine->currentContext()->popScope();
-            engine->currentContext()->popScope();
-            popScopes();
-            if (engine->hasErrorOrException(cr)) {
-                *result = engine->lastErrorValue(cr);
-                return;
-            }
-            if (cr.toBool()) {
-                // condition is true, let's use the value of this alternative
-                if (alternative->value->sourceUsesOuter() && !outerItem) {
-                    // Clone value but without alternatives.
-                    JSSourceValuePtr outerValue = JSSourceValue::create();
-                    outerValue->setFile(value->file());
-                    outerValue->setHasFunctionForm(value->hasFunctionForm());
-                    outerValue->setSourceCode(value->sourceCode());
-                    outerValue->setBaseValue(value->baseValue());
-                    if (value->sourceUsesBase())
-                        outerValue->setSourceUsesBaseFlag();
-                    outerValue->setLocation(value->line(), value->column());
-                    outerItem = Item::create(data->item->pool());
-                    outerItem->setProperty(propertyName->toString(), outerValue);
+        QScriptValue outerScriptValue;
+        for (const JSSourceValue::Alternative &alternative : value->alternatives()) {
+            if (alternative.value->sourceUsesOuter()
+                    && !data->item->outerItem()
+                    && !outerScriptValue.isValid()) {
+                JSSourceValueEvaluationResult sver = evaluateJSSourceValue(value, nullptr);
+                if (sver.hasError) {
+                    *result = sver.scriptValue;
+                    return;
                 }
-                if (overrides.toBool())
-                    value->setIsExclusiveListValue();
-                value = alternative->value.data();
-                break;
+                outerScriptValue = sver.scriptValue;
             }
-        }
-
-        QScriptValue extraScope;
-        if (value->sourceUsesBase()) {
-            QScriptValue baseValue;
-            if (value->baseValue()) {
-                SVConverter converter(scriptClass, object, value->baseValue(), itemOfProperty,
-                                      propertyName, data, &baseValue, sourceValueStack);
-                converter.start();
-            }
-            setupConvenienceProperty(QLatin1String("base"), &extraScope, baseValue);
-        }
-        if (value->sourceUsesOuter() && outerItem) {
-            const QScriptValue v = data->evaluator->property(outerItem, *propertyName);
-            if (engine->hasErrorOrException(v)) {
-                *result = engine->lastErrorValue(v);
+            JSSourceValueEvaluationResult sver = evaluateJSSourceValue(alternative.value.get(),
+                                                                       data->item->outerItem(),
+                                                                       &alternative,
+                                                                       value, &outerScriptValue);
+            if (!sver.tryNextAlternative || sver.hasError) {
+                *result = sver.scriptValue;
                 return;
             }
-            setupConvenienceProperty(QLatin1String("outer"), &extraScope, v);
-        }
-        if (value->sourceUsesOriginal()) {
-            QScriptValue originalValue;
-            if (data->item->propertyDeclaration(propertyName->toString()).isScalar()) {
-                const Item *item = itemOfProperty;
-                while (item->type() == ItemType::ModuleInstance)
-                    item = item->prototype();
-                SVConverter converter(scriptClass, object, item->property(*propertyName), item,
-                                      propertyName, data, &originalValue, sourceValueStack);
-                converter.start();
-            } else {
-                originalValue = engine->newArray(0);
-            }
-            setupConvenienceProperty(QLatin1String("original"), &extraScope, originalValue);
         }
+        *result = evaluateJSSourceValue(value, data->item->outerItem()).scriptValue;
+    }
 
-        pushScope(data->evaluator->fileScope(value->file()));
+    struct JSSourceValueEvaluationResult
+    {
+        QScriptValue scriptValue;
+        bool tryNextAlternative = true;
+        bool hasError = false;
+    };
+
+    void injectErrorLocation(QScriptValue &sv, const CodeLocation &loc)
+    {
+        if (sv.isError() && !engine->lastErrorLocation(sv).isValid())
+            sv = engine->currentContext()->throwError(engine->lastError(sv, loc).toString());
+    }
+
+    JSSourceValueEvaluationResult evaluateJSSourceValue(const JSSourceValue *value, Item *outerItem,
+            const JSSourceValue::Alternative *alternative = nullptr,
+            JSSourceValue *elseCaseValue = nullptr, QScriptValue *outerScriptValue = nullptr)
+    {
+        JSSourceValueEvaluationResult result;
+        QBS_ASSERT(!alternative || value == alternative->value.get(), return result);
+        AutoScopePopper autoScopePopper(this);
+        auto maybeExtraScope = createExtraScope(value, outerItem, outerScriptValue);
+        if (!maybeExtraScope.second) {
+            result.scriptValue = maybeExtraScope.first;
+            result.hasError = true;
+            return result;
+        }
+        const Evaluator::FileContextScopes fileCtxScopes
+                = data->evaluator->fileContextScopes(value->file());
+        if (fileCtxScopes.importScope.isError()) {
+            result.scriptValue = fileCtxScopes.importScope;
+            result.hasError = true;
+            return result;
+        }
+        pushScope(fileCtxScopes.fileScope);
         pushItemScopes(data->item);
-        if (itemOfProperty && itemOfProperty->type() != ItemType::ModuleInstance) {
+        if (itemOfProperty->type() != ItemType::ModuleInstance) {
             // Own properties of module instances must not have the instance itself in the scope.
             pushScope(*object);
         }
         if (value->definingItem())
             pushItemScopes(value->definingItem());
-        pushScope(extraScope);
-        pushScope(data->evaluator->importScope(value->file()));
-        *result = engine->evaluate(value->sourceCodeForEvaluation(), value->file()->filePath(),
-                                   value->line());
-        popScopes();
+        pushScope(maybeExtraScope.first);
+        pushScope(fileCtxScopes.importScope);
+        if (alternative) {
+            QScriptValue sv = engine->evaluate(alternative->condition.value);
+            if (engine->hasErrorOrException(sv)) {
+                result.scriptValue = sv;
+                result.hasError = true;
+                injectErrorLocation(result.scriptValue, alternative->condition.location);
+                return result;
+            }
+            if (sv.toBool()) {
+                // The condition is true. Continue evaluating the value.
+                result.tryNextAlternative = false;
+            } else {
+                // The condition is false. Try the next alternative or the else value.
+                result.tryNextAlternative = true;
+                return result;
+            }
+            sv = engine->evaluate(alternative->overrideListProperties.value);
+            if (engine->hasErrorOrException(sv)) {
+                result.scriptValue = sv;
+                result.hasError = true;
+                injectErrorLocation(result.scriptValue,
+                                    alternative->overrideListProperties.location);
+                return result;
+            }
+            if (sv.toBool())
+                elseCaseValue->setIsExclusiveListValue();
+        }
+        result.scriptValue = engine->evaluate(value->sourceCodeForEvaluation(),
+                                              value->file()->filePath(), value->line());
+        return result;
     }
 
-    void handle(ItemValue *value)
+    void handle(ItemValue *value) override
     {
         *result = data->evaluator->scriptValue(value->item());
         if (!result->isValid())
             qDebug() << "SVConverter returned invalid script value.";
     }
 
-    void handle(VariantValue *variantValue)
+    void handle(VariantValue *variantValue) override
     {
-        *result = scriptClass->engine()->toScriptValue(variantValue->value());
+        *result = engine->toScriptValue(variantValue->value());
     }
 };
 
@@ -296,12 +341,10 @@ enum QueryPropertyType
     QPTParentProperty
 };
 
-EvaluatorScriptClass::EvaluatorScriptClass(ScriptEngine *scriptEngine, Logger &logger)
+EvaluatorScriptClass::EvaluatorScriptClass(ScriptEngine *scriptEngine)
     : QScriptClass(scriptEngine)
-    , m_logger(logger)
     , m_valueCacheEnabled(false)
 {
-    Q_UNUSED(m_logger);
 }
 
 QScriptClass::QueryFlags EvaluatorScriptClass::queryProperty(const QScriptValue &object,
@@ -320,7 +363,7 @@ QScriptClass::QueryFlags EvaluatorScriptClass::queryProperty(const QScriptValue
 
     EvaluationData *const data = attachedPointer<EvaluationData>(object);
     const QString nameString = name.toString();
-    if (nameString == QLatin1String("parent")) {
+    if (nameString == QStringLiteral("parent")) {
         *id = QPTParentProperty;
         m_queryResult.data = data;
         return QScriptClass::HandlesReadAccess;
@@ -341,8 +384,8 @@ QScriptClass::QueryFlags EvaluatorScriptClass::queryItemProperty(const Evaluatio
                                                                  bool ignoreParent)
 {
     for (const Item *item = data->item; item; item = item->prototype()) {
-        m_queryResult.value = item->properties().value(name);
-        if (!m_queryResult.value.isNull()) {
+        m_queryResult.value = item->ownProperty(name);
+        if (m_queryResult.value) {
             m_queryResult.data = data;
             m_queryResult.itemOfProperty = item;
             return HandlesReadAccess;
@@ -356,6 +399,7 @@ QScriptClass::QueryFlags EvaluatorScriptClass::queryItemProperty(const Evaluatio
         parentdata.item = data->item->parent();
         const QueryFlags qf = queryItemProperty(&parentdata, name, true);
         if (qf.testFlag(HandlesReadAccess)) {
+            m_queryResult.foundInParent = true;
             m_queryResult.data = data;
             return qf;
         }
@@ -370,7 +414,7 @@ QString EvaluatorScriptClass::resultToString(const QScriptValue &scriptValue)
 {
     return (scriptValue.isObject()
         ? QLatin1String("[Object: ")
-            + QString::number(scriptValue.objectId(), 16) + QLatin1Char(']')
+            + QString::number(scriptValue.objectId()) + QLatin1Char(']')
         : scriptValue.toVariant().toString());
 }
 
@@ -378,13 +422,13 @@ void EvaluatorScriptClass::collectValuesFromNextChain(const EvaluationData *data
         const QString &propertyName, const ValuePtr &value)
 {
     QScriptValueList lst;
-    QSet<Value *> oldNextChain = m_currentNextChain;
+    Set<Value *> oldNextChain = m_currentNextChain;
     for (ValuePtr next = value; next; next = next->next())
-        m_currentNextChain.insert(next.data());
+        m_currentNextChain.insert(next.get());
 
     for (ValuePtr next = value; next; next = next->next()) {
         QScriptValue v = data->evaluator->property(next->definingItem(), propertyName);
-        ScriptEngine * const se = static_cast<ScriptEngine *>(engine());
+        const auto se = static_cast<const ScriptEngine *>(engine());
         if (se->hasErrorOrException(v)) {
             *result = se->lastErrorValue(v);
             return;
@@ -393,7 +437,7 @@ void EvaluatorScriptClass::collectValuesFromNextChain(const EvaluationData *data
             continue;
         lst << v;
         if (next->type() == Value::JSSourceValueType
-                && next.staticCast<JSSourceValue>()->isExclusiveListValue()) {
+                && std::static_pointer_cast<JSSourceValue>(next)->isExclusiveListValue()) {
             lst = lst.mid(lst.length() - 2);
             break;
         }
@@ -402,11 +446,10 @@ void EvaluatorScriptClass::collectValuesFromNextChain(const EvaluationData *data
 
     *result = engine()->newArray();
     quint32 k = 0;
-    for (int i = 0; i < lst.count(); ++i) {
-        const QScriptValue &v = lst.at(i);
+    for (const QScriptValue &v : qAsConst(lst)) {
         QBS_ASSERT(!v.isError(), continue);
         if (v.isArray()) {
-            const quint32 vlen = v.property(QStringLiteral("length")).toInt32();
+            const quint32 vlen = v.property(StringConstants::lengthProperty()).toInt32();
             for (quint32 j = 0; j < vlen; ++j)
                 result->setProperty(k++, v.property(j));
         } else {
@@ -415,10 +458,11 @@ void EvaluatorScriptClass::collectValuesFromNextChain(const EvaluationData *data
     }
 }
 
-static QString overriddenSourceDirectory(const Item *item)
+static QString overriddenSourceDirectory(const Item *item, const QString &defaultValue)
 {
-    const VariantValuePtr v = item->variantProperty(QLatin1String("_qbs_sourceDir"));
-    return v ? v->value().toString() : QString();
+    const VariantValuePtr v = item->variantProperty
+            (StringConstants::qbsSourceDirPropertyInternal());
+    return v ? v->value().toString() : defaultValue;
 }
 
 static void makeTypeError(const ErrorInfo &error, QScriptValue &v)
@@ -435,18 +479,19 @@ static void makeTypeError(const PropertyDeclaration &decl, const CodeLocation &l
     makeTypeError(error, v);
 }
 
-static void convertToPropertyType(const Item *item, const PropertyDeclaration& decl,
-                                  const Value *value, QScriptValue &v)
+static void convertToPropertyType_impl(const QString &pathPropertiesBaseDir, const Item *item,
+                                       const PropertyDeclaration& decl,
+                                       const CodeLocation &location, QScriptValue &v)
 {
-    if (value->type() == Value::VariantValueType && v.isUndefined() && !decl.isScalar()) {
-        v = v.engine()->newArray(); // QTBUG-51237
-        return;
-    }
-
     if (v.isUndefined() || v.isError())
         return;
     QString srcDir;
-    const CodeLocation &location = value->location();
+    QString actualBaseDir;
+    if (item && !pathPropertiesBaseDir.isEmpty()) {
+        const VariantValueConstPtr itemSourceDir
+                = item->variantProperty(QLatin1String("sourceDirectory"));
+        actualBaseDir = itemSourceDir ? itemSourceDir->value().toString() : pathPropertiesBaseDir;
+    }
     switch (decl.type()) {
     case PropertyDeclaration::UnknownType:
     case PropertyDeclaration::Variant:
@@ -465,9 +510,11 @@ static void convertToPropertyType(const Item *item, const PropertyDeclaration& d
             makeTypeError(decl, location, v);
             break;
         }
-        const QString srcDir = overriddenSourceDirectory(item);
+        const QString srcDir = item ? overriddenSourceDirectory(item, actualBaseDir)
+                                    : pathPropertiesBaseDir;
         if (!srcDir.isEmpty())
-            v = v.engine()->toScriptValue(FileInfo::resolvePath(srcDir, v.toString()));
+            v = v.engine()->toScriptValue(QDir::cleanPath(
+                                              FileInfo::resolvePath(srcDir, v.toString())));
         break;
     }
     case PropertyDeclaration::String:
@@ -475,7 +522,8 @@ static void convertToPropertyType(const Item *item, const PropertyDeclaration& d
             makeTypeError(decl, location, v);
         break;
     case PropertyDeclaration::PathList:
-        srcDir = overriddenSourceDirectory(item);
+        srcDir = item ? overriddenSourceDirectory(item, actualBaseDir)
+                      : pathPropertiesBaseDir;
         // Fall-through.
     case PropertyDeclaration::StringList:
     {
@@ -484,7 +532,7 @@ static void convertToPropertyType(const Item *item, const PropertyDeclaration& d
             x.setProperty(0, v);
             v = x;
         }
-        const quint32 c = v.property(QLatin1String("length")).toUInt32();
+        const quint32 c = v.property(StringConstants::lengthProperty()).toUInt32();
         for (quint32 i = 0; i < c; ++i) {
             QScriptValue elem = v.property(i);
             if (elem.isUndefined()) {
@@ -507,7 +555,8 @@ static void convertToPropertyType(const Item *item, const PropertyDeclaration& d
             }
             if (srcDir.isEmpty())
                 continue;
-            elem = v.engine()->toScriptValue(FileInfo::resolvePath(srcDir, elem.toString()));
+            elem = v.engine()->toScriptValue(
+                        QDir::cleanPath(FileInfo::resolvePath(srcDir, elem.toString())));
             v.setProperty(i, elem);
         }
         break;
@@ -515,29 +564,43 @@ static void convertToPropertyType(const Item *item, const PropertyDeclaration& d
     }
 }
 
+void EvaluatorScriptClass::convertToPropertyType(const PropertyDeclaration &decl,
+                                                 const CodeLocation &loc, QScriptValue &v)
+{
+    convertToPropertyType_impl(QString(), nullptr, decl, loc, v);
+}
+
+void EvaluatorScriptClass::convertToPropertyType(const Item *item, const PropertyDeclaration& decl,
+                                                 const Value *value, QScriptValue &v)
+{
+    if (value->type() == Value::VariantValueType && v.isUndefined() && !decl.isScalar()) {
+        v = v.engine()->newArray(); // QTBUG-51237
+        return;
+    }
+    convertToPropertyType_impl(m_pathPropertiesBaseDir, item, decl, value->location(), v);
+}
+
 class PropertyStackManager
 {
 public:
     PropertyStackManager(const Item *itemOfProperty, const QScriptString &name, const Value *value,
-                         QStack<QualifiedId> &requestedProperties,
+                         std::stack<QualifiedId> &requestedProperties,
                          PropertyDependencies &propertyDependencies)
         : m_requestedProperties(requestedProperties)
     {
         if (value->type() == Value::JSSourceValueType
                 && (itemOfProperty->type() == ItemType::ModuleInstance
-                    || itemOfProperty->type() == ItemType::Module)) {
+                    || itemOfProperty->type() == ItemType::Module
+                    || itemOfProperty->type() == ItemType::Export)) {
             const VariantValueConstPtr varValue
-                    = itemOfProperty->variantProperty(QLatin1String("name"));
-            // QBS_CHECK(varValue);
-            // TODO: Check why the base module sometimes has no name. Code suggests it has to have one.
-            if (varValue) {
-                m_stackUpdate = true;
-                const QualifiedId fullPropName
-                        = QualifiedId::fromString(varValue->value().toString()) << name.toString();
-                if (!requestedProperties.isEmpty())
-                    propertyDependencies[fullPropName].insert(requestedProperties.top());
-                m_requestedProperties.push(fullPropName);
-            }
+                    = itemOfProperty->variantProperty(StringConstants::nameProperty());
+            QBS_ASSERT(varValue, return);
+            m_stackUpdate = true;
+            const QualifiedId fullPropName
+                    = QualifiedId::fromString(varValue->value().toString()) << name.toString();
+            if (!requestedProperties.empty())
+                propertyDependencies[fullPropName].insert(requestedProperties.top());
+            m_requestedProperties.push(fullPropName);
         }
     }
 
@@ -548,20 +611,22 @@ public:
     }
 
 private:
-    QStack<QualifiedId> &m_requestedProperties;
+    std::stack<QualifiedId> &m_requestedProperties;
     bool m_stackUpdate = false;
 };
 
 QScriptValue EvaluatorScriptClass::property(const QScriptValue &object, const QScriptString &name,
                                             uint id)
 {
+    const bool foundInParent = m_queryResult.foundInParent;
     const EvaluationData *data = m_queryResult.data;
     const Item * const itemOfProperty = m_queryResult.itemOfProperty;
-    m_queryResult.data = 0;
-    m_queryResult.itemOfProperty = 0;
+    m_queryResult.foundInParent = false;
+    m_queryResult.data = nullptr;
+    m_queryResult.itemOfProperty = nullptr;
     QBS_ASSERT(data, return QScriptValue());
 
-    const QueryPropertyType qpt = static_cast<QueryPropertyType>(id);
+    const auto qpt = static_cast<QueryPropertyType>(id);
     if (qpt == QPTParentProperty) {
         return data->item->parent()
                 ? data->evaluator->scriptValue(data->item->parent())
@@ -576,7 +641,7 @@ QScriptValue EvaluatorScriptClass::property(const QScriptValue &object, const QS
     if (debugProperties)
         qDebug() << "[SC] property " << name;
 
-    PropertyStackManager propStackmanager(itemOfProperty, name, value.data(),
+    PropertyStackManager propStackmanager(itemOfProperty, name, value.get(),
                                           m_requestedProperties, m_propertyDependencies);
 
     QScriptValue result;
@@ -589,15 +654,18 @@ QScriptValue EvaluatorScriptClass::property(const QScriptValue &object, const QS
         }
     }
 
-    if (value->next() && !m_currentNextChain.contains(value.data())) {
+    if (value->next() && !m_currentNextChain.contains(value.get())) {
         collectValuesFromNextChain(data, &result, name.toString(), value);
     } else {
-        SVConverter converter(this, &object, value, itemOfProperty, &name, data, &result,
-                              &m_sourceValueStack);
+        QScriptValue parentObject;
+        if (foundInParent)
+            parentObject = data->evaluator->scriptValue(data->item->parent());
+        SVConverter converter(this, foundInParent ? &parentObject : &object, value, itemOfProperty,
+                              &name, data, &result);
         converter.start();
 
         const PropertyDeclaration decl = data->item->propertyDeclaration(name.toString());
-        convertToPropertyType(data->item, decl, value.data(), result);
+        convertToPropertyType(data->item, decl, value.get(), result);
     }
 
     if (debugProperties)
@@ -607,6 +675,59 @@ QScriptValue EvaluatorScriptClass::property(const QScriptValue &object, const QS
     return result;
 }
 
+class EvaluatorScriptClassPropertyIterator : public QScriptClassPropertyIterator
+{
+public:
+    EvaluatorScriptClassPropertyIterator(const QScriptValue &object, EvaluationData *data)
+        : QScriptClassPropertyIterator(object), m_it(data->item->properties())
+    {
+    }
+
+    bool hasNext() const override
+    {
+        return m_it.hasNext();
+    }
+
+    void next() override
+    {
+        m_it.next();
+    }
+
+    bool hasPrevious() const override
+    {
+        return m_it.hasPrevious();
+    }
+
+    void previous() override
+    {
+        m_it.previous();
+    }
+
+    void toFront() override
+    {
+        m_it.toFront();
+    }
+
+    void toBack() override
+    {
+        m_it.toBack();
+    }
+
+    QScriptString name() const override
+    {
+        return object().engine()->toStringHandle(m_it.key());
+    }
+
+private:
+    QMapIterator<QString, ValuePtr> m_it;
+};
+
+QScriptClassPropertyIterator *EvaluatorScriptClass::newIterator(const QScriptValue &object)
+{
+    EvaluationData *const data = attachedPointer<EvaluationData>(object);
+    return data ? new EvaluatorScriptClassPropertyIterator(object, data) : nullptr;
+}
+
 void EvaluatorScriptClass::setValueCacheEnabled(bool enabled)
 {
     m_valueCacheEnabled = enabled;
